package ij.gui;

import ij.process.FloatPolygon;
import java.awt.geom.Rectangle2D;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Polygon;

import java.awt.Rectangle;

import java.awt.Shape;

import ij.ImagePlus;

import ij.process.ImageProcessor;
import java.awt.Graphics;

/**
 * <p>Title: </p>
 *
 * <p>Description: </p>
 *
 * <p>Copyright: Copyright (c) 2011</p>
 *
 * <p>Company: </p>
 *
 * @author not attributable
 * @version 1.0
 */
public interface RoiBeans extends RoiConstance{

    /** Set the location of the ROI in image coordinates. */
    public void setLocation(int x, int y);
    public void setImage(ImagePlus imp);
    /** Returns the ImagePlus associated with this ROI, or null. */
    public ImagePlus getImage();
    /** Returns the ID of the image associated with this ROI. */
    public int getImageID();

    public int getType();

    public int getState();

    /** Returns the perimeter length. */
    public double getLength();

    /** Returns Feret's diameter, the greatest distance between
            any two points along the ROI boundary. */
    public double getFeretsDiameter();

    /** Caculates "Feret" (maximum caliper width), "FeretAngle"
            and "MinFeret" (minimum caliper width), "FeretX" and "FeretY". */
    public double[] getFeretValues();

    public Polygon getConvexHull();

    //double getFeretBreadth(Shape shape, double angle, double x1, double y1, double x2, double y2);
    /** Return this selection's bounding rectangle. */
    public Rectangle getBounds();
    /** Return this selection's bounding rectangle. */
    public Rectangle2D.Double getFloatBounds();

    /**
    * @deprecated
    * replaced by getBounds()
    */
    public Rectangle getBoundingRect();
    /** Returns the outline of this selection as a Polygon, or
            null if this is a straight line selection.
            @see ij.process.ImageProcessor#setRoi
            @see ij.process.ImageProcessor#drawPolygon
            @see ij.process.ImageProcessor#fillPolygon
    */
    public Polygon getPolygon();

    public FloatPolygon getFloatPolygon();

    /** Nudge ROI one pixel on arrow key press. */
    public void nudge(int key);
    /** Nudge lower right corner of rectangular and oval ROIs by
            one pixel based on arrow key press. */
    public void nudgeCorner(int key);

    public int getHandleSize();

    public void draw(Graphics g);

    public void drawOverlay(Graphics g);

    public void drawHandle(Graphics g, int x, int y);
    /**
    * @deprecated
    * replaced by drawPixels(ImageProcessor)
    */
    public void drawPixels();

    public void drawPixels(ImageProcessor ip);

    public boolean contains(int x, int y);

    /** Returns a handle number if the specified screen coordinates are
            inside or near a handle, otherwise returns -1. */
    public int isHandle(int sx, int sy);

/** If 'add' is true, adds this selection to the previous one. If 'subtract' is true, subtracts
    it from the previous selection. Called by the IJ.doWand() method, and the makeRectangle(),
    makeOval(), makePolygon() and makeSelection() macro functions. */
    public void update(boolean add, boolean subtract);
    //protected void showStatus();
    /** Always returns null for rectangular Roi's */
    public ImageProcessor getMask();
    /** Returns the angle in degrees between the specified line and a horizontal line. */
    public double getAngle(int x1, int y1, int x2, int y2);


    /** Sets the color used by this ROI to draw its outline. This color, if not null,
     * overrides the global color set by the static setColor() method.
     * @see #getStrokeColor()
     * @see #setStrokeWidth(int)
     * @see ij.ImagePlus#setOverlay(ij.gui.Overlay)
     */
    public void setStrokeColor(Color c);

    /** Returns the the color used to draw the ROI outline or null if the default color is being used.
     * @see #setStrokeColor(Color)
     */
    public Color getStrokeColor();

    /** Sets the color used to fill ROIs when they are in an overlay.
     * @see ij.ImagePlus#setOverlay(ij.gui.Overlay)
     */
    public void setFillColor(Color color);

    /** Returns the color used to fill this ROI, or null if it is not filled.
     * @see #getStrokeColor()
     */
    public Color getFillColor();

    /** Copy the attributes (outline color, fill color, outline width)
            of  'roi2' to the this selection. */
    public void copyAttributes(Roi roi2);

    /**
    * @deprecated
    * replaced by setStrokeColor()
    */
    public void setInstanceColor(Color c);

    /**
    * @deprecated
    * replaced by setStrokeWidth(int)
    */
    public void setLineWidth(int width);

    public void updateWideLine(float width);

/** Set 'nonScalable' true to have TextRois in a display
            list drawn at a fixed location and size. */
    public void setNonScalable(boolean nonScalable);

    /** Sets the width of the line used to draw this ROI.
     * @see #setStrokeColor(Color)
     * @see ij.ImagePlus#setOverlay(ij.gui.Overlay)
     */
    public void setStrokeWidth(float width);


    /** This is a version of setStrokeWidth() that accepts a double argument. */
    public void setStrokeWidth(double width);

    /** Returns the lineWidth. */
    public float getStrokeWidth();

    /** Sets the Stroke used to draw this ROI. */
    public void setStroke(BasicStroke stroke);

    /** Returns the Stroke used to draw this ROI, or null if no Stroke is used. */
    public BasicStroke getStroke();

    //protected BasicStroke getScaledStroke();

    /** Returns the name of this ROI, or null. */
    public String getName();

    /** Sets the name of this ROI. */
    public void setName(String name);

    /** Sets the rounded rectangle corner diameter (pixels). */
    public void setCornerDiameter(int cornerDiameter);

    /** Returns the rounded rectangle corner diameter (pixels). */
    public int getCornerDiameter();

    /** Obsolete; replaced by setCornerDiameter(). */
    public void setRoundRectArcSize(int cornerDiameter);

    /** Obsolete; replaced by getCornerDiameter(). */
    public int getRoundRectArcSize();

    /** Sets the stack position (image number) of this ROI. In an overlay, this
    * ROI is only displayed when the stack is at the specified position.
    * Set to zero to have the ROI displayed on all images in the stack.
    * @see ij.gui.Overlay
    */
    public void setPosition(int n);

    /** Returns the stack position (image number) of this ROI, or
    *  zero if the ROI is not associated with a particular stack image.
    * @see ij.gui.Overlay
    */
    public int getPosition();

    /** Sets the hyperstack position of this ROI. In an overlay, this
    * ROI is only displayed when the hyperstack is at the specified position.
    * @see ij.gui.Overlay
    */
    public void setPosition(int channel, int slice, int frame);

    // Used by the FileSaver and RoiEncoder to save overlay settings. */
    public void setPrototypeOverlay(Overlay overlay);

    // Used by the FileOpener and RoiDecoder to restore overlay settings. */
    public Overlay getPrototypeOverlay();

     public void startPaste(ImagePlus clipboard);

     //void updatePaste();

     public void endPaste();

     public void abortPaste();


    /** Returns the current paste transfer mode, or NOT_PASTING (-1)
            if no paste operation is in progress.
            @see ij.process.Blitter
    */
    public int getPasteMode();


    /** Returns 'true' if this is an area selection. */
    public boolean isArea();

    /** Returns 'true' if this is a line selection. */
    public boolean isLine();

    /** Returns 'true' if this is an ROI primarily used from drawing
            (e.g., TextRoi or Arrow). */
    public boolean isDrawingTool();

    /** Convenience method that converts Roi type to a human-readable form. */
    public String getTypeAsString();

    /** Returns true if this ROI is currently displayed on an image. */
    public boolean isVisible();

    /** Returns true if this is a PolygonRoi that supports sub-pixel resolution. */
    public boolean subPixelResolution();

    /** Returns true if this is a PolygonRoi that supports sub-pixel
            resolution and polygons are drawn on zoomed images offset
            down and to the right by 0.5 pixels.. */
    public boolean getDrawOffset();

    public void setDrawOffset(boolean drawOffset);

    /** Checks whether two rectangles are equal. */
    public boolean equals(Object obj);

    public String toString();

}
